library angular2.src.http.backends.xhr_backend;

import "../interfaces.dart" show ConnectionBackend, Connection;
import "../enums.dart" show ReadyStates, RequestMethods, RequestMethodsMap;
import "../static_request.dart" show Request;
import "../static_response.dart" show Response;
import "../base_response_options.dart"
    show ResponseOptions, BaseResponseOptions;
import "package:angular2/di.dart" show Injectable;
import "browser_xhr.dart" show BrowserXhr;
import "package:angular2/src/facade/async.dart"
    show EventEmitter, ObservableWrapper;
import "package:angular2/src/facade/lang.dart" show isPresent, ENUM_INDEX;

/**
 * Creates connections using `XMLHttpRequest`. Given a fully-qualified
 * request, an `XHRConnection` will immediately create an `XMLHttpRequest` object and send the
 * request.
 *
 * This class would typically not be created or interacted with directly inside applications, though
 * the [MockConnection] may be interacted with in tests.
 */
class XHRConnection implements Connection {
  Request request;
  /**
   * Response [EventEmitter] which emits a single [Response] value on load event of
   * `XMLHttpRequest`.
   */
  EventEmitter response;
  ReadyStates readyState;
  var _xhr;
  // https://github.com/angular/ts2dart/issues/230
  XHRConnection(Request req, BrowserXhr browserXHR,
      [ResponseOptions baseResponseOptions]) {
    // TODO: get rid of this when enum lookups are available in ts2dart

    // https://github.com/angular/ts2dart/issues/221
    var requestMethodsMap = new RequestMethodsMap();
    this.request = req;
    this.response = new EventEmitter();
    this._xhr = browserXHR.build();
    // TODO(jeffbcross): implement error listening/propagation
    this._xhr.open(
        requestMethodsMap.getMethod(ENUM_INDEX(req.method)), req.url);
    this._xhr.addEventListener("load", (_) {
      // responseText is the old-school way of retrieving response (supported by IE8 & 9)

      // response/responseType properties were introduced in XHR Level2 spec (supported by IE10)
      var response = isPresent(this._xhr.response)
          ? this._xhr.response
          : this._xhr.responseText;
      // normalize IE9 bug (http://bugs.jquery.com/ticket/1450)
      var status = identical(this._xhr.status, 1223) ? 204 : this._xhr.status;
      // fix status code when it is 0 (0 status is undocumented).

      // Occurs when accessing file resources or on Android 4.1 stock browser

      // while retrieving files from application cache.
      if (identical(status, 0)) {
        status = response ? 200 : 0;
      }
      var responseOptions = new ResponseOptions(body: response, status: status);
      if (isPresent(baseResponseOptions)) {
        responseOptions = baseResponseOptions.merge(responseOptions);
      }
      ObservableWrapper.callNext(this.response, new Response(responseOptions));
      // TODO(gdi2290): defer complete if array buffer until done
      ObservableWrapper.callReturn(this.response);
    });
    // TODO(jeffbcross): make this more dynamic based on body type
    if (isPresent(req.headers)) {
      req.headers.forEach((value, name) {
        this._xhr.setRequestHeader(name, value);
      });
    }
    this._xhr.send(this.request.text());
  }
  /**
   * Calls abort on the underlying XMLHttpRequest.
   */
  void dispose() {
    this._xhr.abort();
  }
}
/**
 * Creates [XHRConnection] instances.
 *
 * This class would typically not be used by end users, but could be
 * overridden if a different backend implementation should be used,
 * such as in a node backend.
 *
 * #Example
 *
 * ```
 * import {Http, MyNodeBackend, httpInjectables, BaseRequestOptions} from 'angular2/http';
 * @Component({
 *   viewBindings: [
 *     httpInjectables,
 *     bind(Http).toFactory((backend, options) => {
 *       return new Http(backend, options);
 *     }, [MyNodeBackend, BaseRequestOptions])]
 * })
 * class MyComponent {
 *   constructor(http:Http) {
 *     http('people.json').subscribe(res => this.people = res.json());
 *   }
 * }
 * ```
 *
 **/
@Injectable()
class XHRBackend implements ConnectionBackend {
  BrowserXhr _browserXHR;
  ResponseOptions _baseResponseOptions;
  XHRBackend(this._browserXHR, this._baseResponseOptions) {}
  XHRConnection createConnection(Request request) {
    return new XHRConnection(
        request, this._browserXHR, this._baseResponseOptions);
  }
}
